@@ -31,7 +31,7 @@ from pesto.request import Request
 from pesto.wsgiutils import uri_join, make_query
 from pesto.httputils import parse_header
 
-__version__ = '7.3.6'
+__version__ = '7.3.7'
 
 #: Registry for xpath multimethods
 xpath_registry = {}
@@ -550,14 +550,18 @@ class ElementWrapper(object):
             self(selector).fill(value)
 
         for name, value in kwargs.items():
-            path = """
-            .//*[
-                (local-name() = 'input'\
-                 or local-name() = 'textarea'\
-                 or local-name() = 'select')
-                and (@name=$name or @id=$name)
-            ]"""
-            self.find(path, name=name).fill(value)
+            path = ".//*[(local-name() = 'input' "\
+                   "or local-name() = 'textarea' "\
+                   "or local-name() = 'select') "\
+                   "and (@name=$name or @id=$name)]"
+            element = self.find(path, name=name)
+            if len(element) == 0:
+                fields = ', '.join(e.name
+                                   for e in self('css, textarea, select'))
+                raise IndexError(
+                    "Couldn't find a form element named {0!r}. "
+                    "Valid names are {1}".format(name, fields))
+            element.fill(value)
 
         return self
 
@@ -1055,7 +1059,8 @@ class TestAgent(object):
 
         return environ
 
-    def _request(self, environ, follow=True, history=False, check_status=True):
+    def _wsgi_request(self, environ, follow=True, history=False,
+                      check_status=True):
         """
         Low level entry point for making requests to the WSGI application.
 
@@ -1086,7 +1091,7 @@ class TestAgent(object):
 
         if self.options['logger']:
             self.options['logger'].info("%s %s", environ['REQUEST_METHOD'],
-                                        Request(environ).url)
+                                        Request(environ).request_uri)
             if environ['HTTP_COOKIE']:
                 self.options['logger'].debug("Cookie: %s",
                                              environ['HTTP_COOKIE'])
@@ -1101,7 +1106,7 @@ class TestAgent(object):
             self.app,
             Request(environ),
             response,
-            self.cookies,
+            self.cookies.copy(),
             history,
             environ_defaults=self.environ_defaults,
             **self.options
@@ -1122,52 +1127,11 @@ class TestAgent(object):
             return agent.follow_all()
         return agent
 
-    def _get(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-             history=True, check_status=True, method='GET', **kwargs):
-        """
-        Make a GET request to the application and return the response.
-
-        :param PATH_INFO:
-            The path to request from the application. This must be URL encoded.
-            As a convenience, query string data can be appended to this.
-            Example value: ``agent.get('/my%20script?param=1')``
-
-        :param data: Query string data to be appended to the request, must be
-                     either a dict or list of ``(name, value)`` tuples.
-
-        :charset: Encoding used for any unicode values encountered in ``data``
-
-        :follow: If false, redirect responses will not be followed
-
-        :history: If true, a new entry will be added to the history attribute
-                  for the resulting TestAgent object
-
-        :check_status: If true, any non success HTTP status code will result in
-                       an AssertionError being raised
-        """
-        if data is not None:
-            kwargs.setdefault('QUERY_STRING', make_query(data,
-                                                         charset=charset))
-
-        if self.request:
-            PATH_INFO = url_join_same_server(self.request.request_uri,
-                                             PATH_INFO)
-
-        return self._request(
-            self.make_environ(
-                'GET',
-                PATH_INFO,
-                **kwargs
-            ),
-            follow,
-            history,
-            check_status,
-        )
-
-    def _post(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
-              history=True, check_status=True,
-              content_type='application/x-www-form-urlencoded', method='POST',
-              **kwargs):
+    def _request(self, PATH_INFO='/', data=None, charset='UTF-8', follow=True,
+                 history=True, check_status=True,
+                 content_type='application/x-www-form-urlencoded',
+                 method='POST',
+                 **kwargs):
         """
         Make a POST request to the application and return the response.
 
@@ -1194,75 +1158,82 @@ class TestAgent(object):
         :param check_status: If true, any non success HTTP status code will
                              result in an AssertionError being raised
         """
-        if data is None:
-            data = []
 
         if self.request:
             PATH_INFO = url_join_same_server(self.request.request_uri,
                                              PATH_INFO)
+        else:
+            baseurl = '{0}://{1}:{2}'.format(
+                            self.environ_defaults['wsgi.url_scheme'],
+                            self.options['host'],
+                            self.options['port'])
+            PATH_INFO = url_join_same_server(baseurl, PATH_INFO)
+
+        if data is None:
+            envargs = kwargs
+        else:
+            if not isinstance(data, bytes):
+                data = make_query(data, charset=charset).encode('ASCII')
+            envargs = {'CONTENT_TYPE': content_type,
+                       'CONTENT_LENGTH': str(len(data))}
+            envargs.update(kwargs)
 
-        if not isinstance(data, bytes):
-            data = make_query(data, charset=charset).encode('ASCII')
         wsgi_input = BytesIO(data)
         wsgi_input.seek(0)
 
-        return self._request(
+        return self._wsgi_request(
             self.make_environ(
                 method,
                 PATH_INFO,
-                CONTENT_TYPE=content_type,
-                CONTENT_LENGTH=str(len(data)),
                 wsgi_input=wsgi_input,
-                **kwargs
+                **envargs
             ),
             follow,
             history,
             check_status,
         )
 
-    get = _get
-    post = _post
+    def get(self, PATH_INFO='/', data=None, charset='UTF-8', *args, **kwargs):
+        kwargs.setdefault('method', 'GET')
+        if data is not None:
+            kwargs.setdefault('QUERY_STRING', make_query(data,
+                                                         charset=charset))
+            data = None
+        return self._request(PATH_INFO, data, charset, *args, **kwargs)
 
     def head(self, *args, **kwargs):
-        return self._get(method='HEAD', *args, **kwargs)
+        return self.get(method='HEAD', *args, **kwargs)
+
+    def post(self, *args, **kwargs):
+        return self._request(method='POST', *args, **kwargs)
 
     def delete(self, *args, **kwargs):
-        return self._get(method='DELETE', *args, **kwargs)
+        return self._request(method='DELETE', *args, **kwargs)
 
     def put(self, *args, **kwargs):
-        return self._post(method='PUT', *args, **kwargs)
+        return self._request(method='PUT', *args, **kwargs)
 
     def patch(self, *args, **kwargs):
-        return self._post(method='PATCH', *args, **kwargs)
+        return self._request(method='PATCH', *args, **kwargs)
 
-    def post_multipart(self, PATH_INFO='/', data=None, files=None,
-                       charset='UTF-8', follow=True, history=True,
-                       check_status=True, **kwargs):
+    def post_multipart(self, PATH_INFO='/', data=None, charset='UTF-8',
+                       files=None, *args, **kwargs):
         """
         POST a request to the given URI using multipart/form-data encoding.
 
-        PATH_INFO
-            The path to request from the application. This must be a URL
-            encoded string.
+        :param PATH_INFO: The path to request from the application. This must
+                          be a URL encoded string.
 
-        data
-            POST data to be sent to the application, must be either a
-            dict or list of ``(name, value)`` tuples.
+        :param data: POST data to be sent to the application, must be either a
+                     dict or list of ``(name, value)`` tuples.
 
-        files
-            list of ``(name, filename, content_type, data)`` tuples. ``data``
-            may be either a byte string, iterator or file-like object.
-
-        history
-            If true, a new entry will be added to the history attribute for the
-            resulting TestAgent object
-
-        check_status
-            If true, any non success HTTP status code will result in an
-            AssertionError being raised
+        :param charset: Encoding used for any unicode values encountered in
+                        ``data``
 
+        :param files: list of ``(name, filename, content_type, data)`` tuples.
+                      ``data`` may be either a byte string, iterator or
+                      file-like object.
         """
-
         if data is None:
             data = {}
 
@@ -1274,10 +1245,6 @@ class TestAgent(object):
         if files is None:
             files = []
 
-        if self.request:
-            PATH_INFO = url_join_same_server(self.request.request_uri,
-                                             PATH_INFO)
-
         boundary = b'----------------------------------------BoUnDaRyVaLuE'
 
         def add_headers(key, value):
@@ -1330,22 +1297,14 @@ class TestAgent(object):
             post_data.write(b'--' + boundary)
         post_data.write(b'--' + CRLF)
         length = post_data.tell()
-        post_data.seek(0)
+        kwargs.setdefault('method', 'POST')
         kwargs.setdefault('CONTENT_LENGTH', str(length))
-        return self._request(
-            self.make_environ(
-                'POST',
-                PATH_INFO,
-                CONTENT_TYPE=unicode_to_environ(
-                                'multipart/form-data; boundary=' +
-                                boundary.decode('ascii')),
-                wsgi_input=post_data,
-                **kwargs
-            ),
-            follow=follow,
-            history=history,
-            check_status=check_status,
-        )
+        kwargs.setdefault('CONTENT_TYPE', unicode_to_environ(
+                                            'multipart/form-data; boundary=' +
+                                            boundary.decode('ascii')))
+
+        return self._request(PATH_INFO, post_data.getvalue(),
+                             charset=charset, **kwargs)
 
     def reload(self, follow=True, check_status=True):
         """
@@ -1366,7 +1325,8 @@ class TestAgent(object):
             wsgi_input = wsgi_input.input
         wsgi_input.seek(0)
         env['wsgi.input'] = wsgi_input
-        agent = self._request(env, follow=follow, check_status=check_status)
+        agent = self._wsgi_request(env, follow=follow,
+                                   check_status=check_status)
 
         for src, dst in zip(
             self.find('//input|//textarea|//option'),
@@ -1771,12 +1731,28 @@ def url_join_same_server(baseurl, url):
     """
     url = uri_join(baseurl, url)
     url = urlparse(url)
-    if urlparse(baseurl)[:2] != url[:2]:
+    baseurl = urlparse(baseurl)
+    if normalize_host(baseurl.scheme, baseurl.netloc) != \
+            normalize_host(url.scheme, url.netloc):
         raise ValueError("URI links to another server: %s" %
                          (urlunparse(url),))
     return urlunparse(('', '') + url[2:])
 
 
+def normalize_host(scheme, host):
+    """
+    Normalize the host part of the URL
+    """
+    host, _, port = host.partition(':')
+    if port == '80' and scheme in ('http', None):
+        port = ''
+    if port == '443' and scheme == 'https':
+        port = ''
+    if port:
+        return '{0}:{1}'.format(host, port)
+    return host
+
+
 def parse_cookies(response):
     """
     Return a ``Cookie.BaseCookie`` object populated from cookies parsed from
